@@ -20,22 +20,16 @@
 
 #include <lua.hpp>
 #include "gringo/lua.hh"
-#include "clingo_lib.hh"
+#include "clingocontrol.hh"
 
 namespace {
 
-Gringo::Control *newControl(int argc, char const **argv) {
-    return new ClingoLib(argc, argv);
-}
-
-void freeControl(Gringo::Control *ctl) {
-    if (ctl) { delete ctl; }
-}
+DefaultGringoModule g_module;
 
 } // namespace
 
 extern "C" int luaopen_gringo(lua_State *L) {
-    try                             { Gringo::Lua::initlib(L, newControl, freeControl); }
+    try                             { Gringo::Lua::initlib(L, g_module); }
     catch (std::exception const &e) { luaL_error(L, e.what()); }
     catch (...)                     { luaL_error(L, "unknown error"); }
     return 1;
