@@ -20,24 +20,26 @@
 
 #include "Python.h"
 #include "gringo/python.hh"
-#include "clingo_lib.hh"
+#include "clingocontrol.hh"
 
 namespace {
 
-Gringo::Control *newControl(int argc, char const **argv) {
-    return new ClingoLib(argc, argv);
-}
-
-void freeControl(Gringo::Control *ctl) {
-    if (ctl) { delete ctl; }
-}
+DefaultGringoModule g_module;
 
 } // namespace
 
+#if PY_MAJOR_VERSION >= 3
+#define INITRETURN return (PyObject *)
+#define INITFAIL return nullptr;
+PyMODINIT_FUNC PyInit_gringo() {
+#else
+#define INITRETURN
+#define INITFAIL return;
 PyMODINIT_FUNC initgringo() {
-    try                             { Gringo::Python::initlib(newControl, freeControl); }
-    catch (std::bad_alloc const &e) { PyErr_SetString(PyExc_MemoryError, e.what()); }
-    catch (std::exception const &e) { PyErr_SetString(PyExc_RuntimeError, e.what()); }
-    catch (...)                     { PyErr_SetString(PyExc_RuntimeError, "unknown error"); }
+#endif
+    try                             { INITRETURN Gringo::Python::initlib(g_module); }
+    catch (std::bad_alloc const &e) { PyErr_SetString(PyExc_MemoryError, e.what()); INITFAIL }
+    catch (std::exception const &e) { PyErr_SetString(PyExc_RuntimeError, e.what()); INITFAIL }
+    catch (...)                     { PyErr_SetString(PyExc_RuntimeError, "unknown error"); INITFAIL }
 }
 
