@@ -20,8 +20,11 @@
 // }}}
 
 #include "clingocontrol.hh"
+#include "clasp/solver.h"
+#include <program_opts/typed_value.h>
+#include <program_opts/application.h>
 
-// {{{ definition of ClingoLpOutput
+// {{{1 definition of ClingoLpOutput
 
 void ClingoLpOutput::addBody(const LitVec& body) {
     for (auto x : body) {
@@ -95,12 +98,12 @@ bool &ClingoLpOutput::disposeMinimize() {
     return disposeMinimize_;
 }
 
-// }}}
-// {{{ definition of ClingoControl
+// {{{1 definition of ClingoControl
 
 #define LOG if (verbose_) std::cerr
-ClingoControl::ClingoControl(bool clingoMode, Clasp::ClaspFacade *clasp, Clasp::Cli::ClaspCliConfig &claspConfig, PostGroundFunc pgf, PreSolveFunc psf)
-    : clasp(clasp)
+ClingoControl::ClingoControl(Gringo::GringoModule &module, bool clingoMode, Clasp::ClaspFacade *clasp, Clasp::Cli::ClaspCliConfig &claspConfig, PostGroundFunc pgf, PreSolveFunc psf)
+    : scripts(module)
+    , clasp(clasp)
     , claspConfig_(claspConfig)
     , pgf_(pgf)
     , psf_(psf)
@@ -116,12 +119,10 @@ void ClingoControl::parse_() {
 
 void ClingoControl::parse(const StringSeq& files, const ClingoOptions& opts, Clasp::Asp::LogicProgram* claspOut, bool addStdIn) {
     using namespace Gringo;
-    if (opts.wNoRedef)        { message_printer()->disable(W_DEFINE_REDEFINTION); }
-    if (opts.wNoCycle)        { message_printer()->disable(W_DEFINE_CYCLIC);  }
-    if (opts.wNoTermUndef)    { message_printer()->disable(W_TERM_UNDEFINED); }
-    if (opts.wNoAtomUndef)    { message_printer()->disable(W_ATOM_UNDEFINED); }
-    if (opts.wNoNonMonotone)  { message_printer()->disable(W_NONMONOTONE_AGGREGATE); }
-    if (opts.wNoFileIncluded) { message_printer()->disable(W_FILE_INCLUDED); }
+    if (opts.wNoOperationUndefined) { message_printer()->disable(W_OPERATION_UNDEFINED); }
+    if (opts.wNoAtomUndef)          { message_printer()->disable(W_ATOM_UNDEFINED); }
+    if (opts.wNoVariableUnbounded)  { message_printer()->disable(W_VARIABLE_UNBOUNDED); }
+    if (opts.wNoFileIncluded)       { message_printer()->disable(W_FILE_INCLUDED); }
     verbose_ = opts.verbose;
     Output::OutputPredicates outPreds;
     for (auto &x : opts.foobar) {
@@ -135,7 +136,7 @@ void ClingoControl::parse(const StringSeq& files, const ClingoOptions& opts, Cla
         else          { lpOut.reset(new Output::PlainLparseOutputter(std::cout)); }
         out.reset(new Output::OutputBase(std::move(outPreds), *lpOut, opts.lparseDebug));
     }
-    pb = make_unique<Input::NongroundProgramBuilder>(scripts, prg, *out, defs);
+    pb = make_unique<Input::NongroundProgramBuilder>(scripts, prg, *out, defs, opts.rewriteMinimize);
     parser = make_unique<Input::NonGroundParser>(*pb);
     for (auto &x : opts.defines) {
         LOG << "define: " << x << std::endl;
@@ -149,19 +150,17 @@ void ClingoControl::parse(const StringSeq& files, const ClingoOptions& opts, Cla
         LOG << "reading from stdin" << std::endl;
         parser->pushFile("-");
     }
+    updateDomains = opts.updateDomains;
     parse_();
 }
 
-bool ClingoControl::ok() {
-    return !clingoMode_ || (clasp->program() && clasp->program()->ok());
-}
-
 bool ClingoControl::update() {
-    if (clingoMode_ && (clasp->program()->frozen() || configUpdate_)) {
+    if (clingoMode_) {
         clasp->update(configUpdate_);
         configUpdate_ = false;
+        return clasp->ok();
     }
-    return ok();
+    return true;
 }
 
 void ClingoControl::ground(Gringo::Control::GroundVec const &parts, Gringo::Any &&context) {
@@ -176,7 +175,7 @@ void ClingoControl::ground(Gringo::Control::GroundVec const &parts, Gringo::Any
         }
         parsed = false;
     }
-    if (!grounded) {
+    else if (!grounded) {
         if (incremental) { out->incremental(); }
         grounded = true;
     }
@@ -217,9 +216,12 @@ void ClingoControl::onFinish(Clasp::ClaspFacade::Result ret) {
 }
 Gringo::Value ClingoControl::getConst(std::string const &name) {
     auto ret = defs.defs().find(name);
-    return ret != defs.defs().end()
-        ? std::get<2>(ret->second)->eval()
-        : Gringo::Value();
+    if (ret != defs.defs().end()) {
+        bool undefined = false;
+        Gringo::Value val = std::get<2>(ret->second)->eval(undefined);
+        if (!undefined) { return val; }
+    }
+    return Gringo::Value();
 }
 void ClingoControl::add(std::string const &name, Gringo::FWStringVec const &params, std::string const &part) {
     Gringo::Location loc("<block>", 1, 1, "<block>", 1, 1);
@@ -284,7 +286,8 @@ Gringo::ConfigProxy &ClingoControl::getConf() {
 Gringo::SolveIter *ClingoControl::solveIter(Assumptions &&ass) {
     if (!clingoMode_) { throw std::runtime_error("solveIter is not supported in gringo gringo mode"); }
 #if WITH_THREADS
-    prepare_(nullptr, nullptr, std::move(ass));
+    prepare_(nullptr, nullptr);
+    clasp->assume(toClaspAssumptions(std::move(ass)));
     solveIter_ = Gringo::make_unique<ClingoSolveIter>(clasp->startSolveAsync(), static_cast<Clasp::Asp::LogicProgram&>(*clasp->program()), *out, clasp->ctx);
     return solveIter_.get();
 #else
@@ -295,8 +298,9 @@ Gringo::SolveIter *ClingoControl::solveIter(Assumptions &&ass) {
 Gringo::SolveFuture *ClingoControl::solveAsync(ModelHandler mh, FinishHandler fh, Assumptions &&ass) {
     if (!clingoMode_) { throw std::runtime_error("solveAsync is not supported in gringo gringo mode"); }
 #if WITH_THREADS
-    prepare_(mh, fh, std::move(ass));
-    solveFuture_ = Gringo::make_unique<ClingoSolveFuture>(clasp->solveAsync());
+    prepare_(mh, fh);
+    clasp->assume(toClaspAssumptions(std::move(ass)));
+    solveFuture_ = Gringo::make_unique<ClingoSolveFuture>(clasp->solveAsync(nullptr));
     return solveFuture_.get();
 #else
     (void)mh;
@@ -308,10 +312,9 @@ Gringo::SolveFuture *ClingoControl::solveAsync(ModelHandler mh, FinishHandler fh
 bool ClingoControl::blocked() {
     return clasp->solving();
 }
-bool ClingoControl::prepare_(Gringo::Control::ModelHandler mh, Gringo::Control::FinishHandler fh, Gringo::Control::Assumptions &&ass) {
-    if (!update()) { return false; }
+void ClingoControl::prepare_(Gringo::Control::ModelHandler mh, Gringo::Control::FinishHandler fh) {
     grounded = false;
-    out->finish();
+    if (update()) { out->finish(); }
     if (clingoMode_) {
 #if WITH_THREADS
         solveIter_    = nullptr;
@@ -321,29 +324,81 @@ bool ClingoControl::prepare_(Gringo::Control::ModelHandler mh, Gringo::Control::
         modelHandler  = mh;
         Clasp::ProgramBuilder *prg = clasp->program();
         if (lpOut && lpOut->disposeMinimize()) { prg->disposeMinimizeConstraint(); }
-        if (!prg->ok() || (pgf_ && !pgf_(*prg))) { return false; }
-        if (!clasp->prepare(enableEnumAssupmption_ ? Clasp::ClaspFacade::enum_volatile : Clasp::ClaspFacade::enum_static) && (!psf_ || psf_(*clasp))) { return false; }
-        for (auto &x : ass) {
-            auto atm = out->find2(x.first);
-            if (atm && atm->second.hasUid()) {
-                Clasp::Literal lit = static_cast<Clasp::Asp::LogicProgram*>(prg)->getLiteral(atm->second.uid());
-                clasp->assume(x.second ? lit : ~lit);
-            }
-            else if (x.second) {
-                Clasp::Literal lit = static_cast<Clasp::Asp::LogicProgram*>(prg)->getLiteral(1);
-                clasp->assume(lit);
-                break;
-            }
+        if (pgf_) { pgf_(*prg); }
+        clasp->prepare(enableEnumAssupmption_ ? Clasp::ClaspFacade::enum_volatile : Clasp::ClaspFacade::enum_static);
+        if (psf_) {  psf_(*clasp);}
+    }
+}
+
+Clasp::LitVec ClingoControl::toClaspAssumptions(Gringo::Control::Assumptions &&ass) const {
+    Clasp::LitVec outAss;
+    if (!clingoMode_ || !clasp->program()) { return outAss; }
+    const Clasp::Asp::LogicProgram* prg = static_cast<const Clasp::Asp::LogicProgram*>(clasp->program());
+    for (auto &x : ass) {
+        auto atm = out->find2(x.first);
+        if (atm && atm->second.hasUid()) {
+            Clasp::Literal lit = prg->getLiteral(atm->second.uid());
+            outAss.push_back(x.second ? lit : ~lit);
+        }
+        else if (x.second) {
+            Clasp::Literal lit = prg->getLiteral(1);
+            outAss.push_back(lit);
+            break;
         }
     }
-    else if (!ass.empty()) {
-        std::cerr << "warning: the lparse format does not support assumptions" << std::endl;
-    }
-    return true;
+    return outAss;
 }
+
 Gringo::SolveResult ClingoControl::solve(ModelHandler h, Assumptions &&ass) {
-    if (!prepare_(h, nullptr, std::move(ass))) { return convert(clasp->result()); }
-    return clingoMode_ ? convert(clasp->solve()) : Gringo::SolveResult::UNKNOWN;
+    prepare_(h, nullptr);
+    if (clingoMode_) {
+        clasp->assume(toClaspAssumptions(std::move(ass)));
+        auto ret = convert(clasp->solve(nullptr));
+        if (updateDomains) {
+            Clasp::Asp::LogicProgram &prg = static_cast<Clasp::Asp::LogicProgram&>(*clasp->program());
+            Clasp::Solver &solver = *clasp->ctx.solver(0);
+            // TODO: would be nice to have this one in the statistics output
+            unsigned facts = 0;
+            unsigned deleted = 0;
+            for (auto &x : out->domains) {
+                unsigned offset = 0;
+                x.second.indices.clear();
+                x.second.fullIndices.clear();
+                x.second.exports.exports.erase(std::remove_if(x.second.exports.begin(), x.second.exports.end(), [&](Gringo::PredicateDomain::element_type &y) -> bool {
+                    y.second.generation(offset);
+                    if (y.second.hasUid() && !prg.isExternal(y.second.uid())) {
+                        Clasp::Literal lit = prg.getLiteral(prg.getEqAtom(y.second.uid())); 
+                        if (solver.isTrue(lit)) {
+                            // NOTE: externals cannot become facts here
+                            //       because they might get new definitions while grounding
+                            //       because there is no distinction between true and weak true
+                            //       these definitions might be skipped if a weak true external 
+                            //       is made a fact here
+                            if (!y.second.fact(false)) { ++facts; }
+                            y.second.setFact(true);
+                        }
+                        else if (solver.isFalse(lit)) {
+                            if (offset < x.second.exports.incOffset)  { --x.second.exports.incOffset; }
+                            if (offset < x.second.exports.showOffset) { --x.second.exports.showOffset; }
+                            x.second.domain.erase(y.first);
+                            ++deleted;
+                            return true;
+                        }
+                    }
+                    ++offset;
+                    return false;
+                }), x.second.exports.end());
+                x.second.exports.generation_     = 0;
+                x.second.exports.nextGeneration_ = x.second.exports.size();
+            }
+            LOG << facts << " atom" << (facts == 1 ? "" : "s") << " became facts" << std::endl;
+            LOG << deleted << " atom" << (deleted == 1 ? "" : "s") << " deleted" << std::endl;
+        }
+        return ret;
+    }
+    else {
+        return Gringo::SolveResult::UNKNOWN;
+    }
 }
 std::string ClingoControl::str() {
     return "[object:IncrementalControl]";
@@ -352,7 +407,7 @@ void ClingoControl::assignExternal(Gringo::Value ext, Gringo::TruthValue val) {
     if (update()) {
         Gringo::PredicateDomain::element_type *atm = out->find2(ext);
         if (atm && atm->second.hasUid()) {
-            out->external(*atm, val);
+            out->assignExternal(*atm, val);
         }
     }
 }
@@ -367,8 +422,130 @@ bool ClingoControl::useEnumAssumption() {
     return enableEnumAssupmption_;
 }
 
-// }}}
-// {{{ definition of ClingoStatistics
+Gringo::DomainProxy &ClingoControl::getDomain() {
+    if (clingoMode_) { return *this; }
+    else {
+        throw std::runtime_error("domain introspection only supported in clingo mode");
+    }
+}
+
+namespace {
+
+static bool skipDomain(Gringo::FWSignature sig) {
+    return (strncmp((*(*sig).name()).c_str(), "#", 1) == 0);
+}
+
+struct ClingoDomainElement : Gringo::DomainProxy::Element {
+    using ElemIt = Gringo::PredicateDomain::element_vec::iterator;
+    using DomIt = Gringo::PredDomMap::iterator;
+    ClingoDomainElement(Gringo::Output::OutputBase &out, Clasp::Asp::LogicProgram &prg, DomIt const &domIt, ElemIt const &elemIt, bool advanceDom = true)
+    : out(out)
+    , prg(prg)
+    , domIt(domIt)
+    , elemIt(elemIt)
+    , advanceDom(advanceDom) {
+        assert(domIt != out.domains.end() && elemIt != domIt->second.exports.end());
+    }
+    Gringo::Value atom() const { 
+        return elemIt->get().first;
+    }
+    bool fact() const {
+        return elemIt->get().second.fact(false);
+    }
+    bool external() const {
+        return 
+            elemIt->get().second.hasUid() && 
+            elemIt->get().second.isExternal() &&
+            prg.isExternal(elemIt->get().second.uid());
+    }
+    Gringo::DomainProxy::ElementPtr next() {
+        auto domIe  = out.domains.end();
+        auto domIt  = this->domIt;
+        auto elemIt = this->elemIt;
+        if (domIt != domIe) {
+            assert(elemIt != domIt->second.exports.end());
+            ++elemIt;
+            if (elemIt != domIt->second.exports.end()) {
+                return Gringo::make_unique<ClingoDomainElement>(out, prg, domIt, elemIt, advanceDom);
+            }
+            if (advanceDom) {
+                for (++domIt; domIt != domIe; ++domIt) {
+                    if (!skipDomain(domIt->first) && !domIt->second.exports.exports.empty()) {
+                        elemIt = domIt->second.exports.begin();
+                        return Gringo::make_unique<ClingoDomainElement>(out, prg, domIt, elemIt, advanceDom);
+                    }
+                }
+            }
+        }
+        return nullptr;
+    }
+    bool valid() const {
+        return domIt != out.domains.end();
+    }
+    Gringo::Output::OutputBase &out;
+    Clasp::Asp::LogicProgram &prg;
+    DomIt domIt;
+    ElemIt elemIt;
+    bool advanceDom;
+};
+
+} // namespace
+
+std::vector<Gringo::FWSignature> ClingoControl::signatures() const {
+    std::vector<Gringo::FWSignature> ret;
+    for (auto &dom : out->domains) {
+        if (!skipDomain(dom.first)) {
+            ret.emplace_back(dom.first);
+        }
+    }
+    return ret;
+}
+
+Gringo::DomainProxy::ElementPtr ClingoControl::iter(Gringo::Signature const &sig) const {
+    auto it = out->domains.find(sig);
+    if (it != out->domains.end()) {
+        auto jt = it->second.exports.begin();
+        if (jt != it->second.exports.end()) {
+            return Gringo::make_unique<ClingoDomainElement>(*out, static_cast<Clasp::Asp::LogicProgram&>(*clasp->program()), it, jt, false);
+        }
+    }
+    return nullptr;
+}
+
+Gringo::DomainProxy::ElementPtr ClingoControl::iter() const {
+    for (auto it = out->domains.begin(), ie = out->domains.end(); it != ie; ++it) {
+        if (!skipDomain(it->first) && !it->second.exports.exports.empty()) {
+            return Gringo::make_unique<ClingoDomainElement>(*out, static_cast<Clasp::Asp::LogicProgram&>(*clasp->program()), it, it->second.exports.begin());
+        }
+    }
+    return nullptr;
+}
+
+Gringo::DomainProxy::ElementPtr ClingoControl::lookup(Gringo::Value const &atom) const {
+    if (atom.hasSig()) {
+        auto it = out->domains.find(atom.sig());
+        if (it != out->domains.end()) {
+            auto jt = it->second.domain.find(atom);
+            if (jt != it->second.domain.end()) {
+                auto kt = it->second.exports.begin() + jt->second.generation();
+                return Gringo::make_unique<ClingoDomainElement>(*out, static_cast<Clasp::Asp::LogicProgram&>(*clasp->program()), it, kt);
+            }
+        }
+    }
+    return nullptr;
+}
+
+size_t ClingoControl::length() const {
+    size_t ret = 0;
+    for (auto &dom : out->domains) {
+        if (!skipDomain(dom.first)) {
+            ret += dom.second.exports.size();
+        }
+    }
+    return ret;
+}
+
+// {{{1 definition of ClingoStatistics
 
 Gringo::Statistics::Quantity ClingoStatistics::getStat(char const* key) const {
     if (!clasp) { return std::numeric_limits<double>::quiet_NaN(); }
@@ -386,8 +563,7 @@ char const *ClingoStatistics::getKeys(char const* key) const {
     return clasp->getKeys(key);
 }
 
-// }}}
-// {{{ definition of ClingoSolveIter
+// {{{1 definition of ClingoSolveIter
 
 #if WITH_THREADS
 ClingoSolveIter::ClingoSolveIter(Clasp::ClaspFacade::AsyncResult const &future, Clasp::Asp::LogicProgram const &lp, Gringo::Output::OutputBase const &out, Clasp::SharedContext const &ctx) 
@@ -408,8 +584,7 @@ Gringo::SolveResult ClingoSolveIter::get() {
 ClingoSolveIter::~ClingoSolveIter() = default;
 #endif
 
-// }}}
-// {{{ definition of ClingoSolveFuture
+// {{{1 definition of ClingoSolveFuture
 
 Gringo::SolveResult convert(Clasp::ClaspFacade::Result res) {
     switch (res) {
@@ -440,8 +615,98 @@ bool ClingoSolveFuture::wait(double timeout) {
     }
     return true;
 }
-void ClingoSolveFuture::interrupt() { future.cancel(); }
+void ClingoSolveFuture::cancel() { future.cancel(); }
 ClingoSolveFuture::~ClingoSolveFuture() { }
 #endif
 
-// }}}
+// {{{1 declaration of ClingoLib
+
+ClingoLib::ClingoLib(Gringo::GringoModule &module, int argc, char const **argv) 
+        : ClingoControl(module, true, &clasp_, claspConfig_, nullptr, nullptr) { 
+    using namespace ProgramOptions;
+    OptionContext allOpts("<pyclingo>");
+    initOptions(allOpts);
+    ParsedValues values = parseCommandLine(argc, const_cast<char**>(argv), allOpts, false, parsePositional);
+    ParsedOptions parsed;
+    parsed.assign(values);
+    allOpts.assignDefaults(parsed);
+    claspConfig_.finalize(parsed, Clasp::Problem_t::ASP, true);
+    clasp_.ctx.setEventHandler(this);
+    Clasp::Asp::LogicProgram* lp = &clasp_.startAsp(claspConfig_, true);
+    incremental = true;
+    parse({}, grOpts_, lp, false);
+}
+
+
+static bool parseConst(const std::string& str, std::vector<std::string>& out) {
+    out.push_back(str);
+    return true;
+}
+
+void ClingoLib::initOptions(ProgramOptions::OptionContext& root) {
+    using namespace ProgramOptions;
+    grOpts_.defines.clear();
+    grOpts_.verbose = false;
+    OptionGroup gringo("Gringo Options");
+    grOpts_.text = false;
+    gringo.addOptions()
+        ("verbose,V"                , flag(grOpts_.verbose = false), "Enable verbose output")
+        ("const,c"                  , storeTo(grOpts_.defines, parseConst)->composing()->arg("<id>=<term>"), "Replace term occurences of <id> with <term>")
+        ("lparse-rewrite"           , flag(grOpts_.lpRewrite = false), "Use together with --text to inspect lparse rewriting")
+        ("lparse-debug"             , storeTo(grOpts_.lparseDebug = Gringo::Output::LparseDebug::NONE, values<Gringo::Output::LparseDebug>()
+          ("none"  , Gringo::Output::LparseDebug::NONE)
+          ("plain" , Gringo::Output::LparseDebug::PLAIN)
+          ("lparse", Gringo::Output::LparseDebug::LPARSE)
+          ("all"   , Gringo::Output::LparseDebug::ALL)), "Debug information during lparse rewriting:\n"
+         "      none  : no additional info\n"
+         "      plain : print rules as in plain output (prefix %%)\n"
+         "      lparse: print rules as in lparse output (prefix %%%%)\n"
+         "      all   : combines plain and lparse\n")
+        ("update-domains"           , flag(grOpts_.updateDomains = false), "Use top level assignment to cleanup domains")
+        ("warn,W"                   , storeTo(grOpts_, parseWarning)->arg("<warn>")->composing(), "Enable/disable warnings:\n"
+         "      [no-]atom-undefined:        a :- b.\n"
+         "      [no-]define-cyclic:         #const a=b. #const b=a.\n"
+         "      [no-]define-redefinition:   #const a=1. #const a=2.\n"
+         "      [no-]file-included:         #include \"a.lp\". #include \"a.lp\".\n"
+         "      [no-]script-failure:        error in lua or python code\n"
+         "      [no-]operation-undefined:   p(1/0).\n"
+         "      [no-]variable-unbounded:    $x > 10.\n")
+        ("rewrite-minimize"         , flag(grOpts_.rewriteMinimize = false), "Rewrite minimize constraints into rules")
+        ;
+    root.add(gringo);
+    claspConfig_.addOptions(root);
+}
+
+bool ClingoLib::onModel(Clasp::Solver const&, Clasp::Model const& m) {
+    return ClingoControl::onModel(m);
+}
+void ClingoLib::onEvent(Clasp::Event const& ev) {
+#if WITH_THREADS
+    Clasp::ClaspFacade::StepReady const *r = Clasp::event_cast<Clasp::ClaspFacade::StepReady>(ev);
+    if (r && finishHandler) { onFinish(r->summary->result); }
+#endif
+    const Clasp::LogEvent* log = Clasp::event_cast<Clasp::LogEvent>(ev);
+    if (log && log->isWarning()) {
+        fflush(stdout);
+        fprintf(stderr, "*** %-5s: (%s): %s\n", "Warn", "pyclingo", log->msg);
+        fflush(stderr);
+    }
+}
+bool ClingoLib::parsePositional(const std::string& t, std::string& out) {
+    int num;
+    if (bk_lib::string_cast(t, num)) { 
+        out = "number";
+        return true;
+    }
+    return false;
+}
+ClingoLib::~ClingoLib() {
+    // TODO: can be removed after bennies next update...
+#if WITH_THREADS
+    solveIter_   = nullptr;
+    solveFuture_ = nullptr;
+#endif
+    clasp_.shutdown();
+}
+
+// }}}1
