@@ -18,6 +18,15 @@
 
 // }}}
 
+#ifndef _GNU_SOURCE
+#  define _GNU_SOURCE
+#endif
+#include <cstdlib>
+#ifdef __USE_GNU
+#  include <libgen.h>
+#  include <sys/types.h>
+#  include <sys/stat.h>
+#endif
 #include "gringo/input/nongroundparser.hh"
 #include "gringo/lexerstate.hh"
 #include "gringo/value.hh"
@@ -30,8 +39,112 @@
 #include <vector>
 #include <algorithm>
 
+
 namespace Gringo { namespace Input {
 
+namespace {
+
+struct Free {
+    void operator ()(char *ptr) { free(ptr); }
+};
+
+template <typename T>
+void report_included(T const &loc, std::string const &filename) {
+    GRINGO_REPORT(W_FILE_INCLUDED) << loc << ": warning: already included file:\n"
+        << "  " << filename << "\n";
+}
+
+template <typename T>
+void report_not_found(T const &loc, std::string const &filename) {
+    GRINGO_REPORT(ERROR) << loc << ": error: file could not be opened:\n"
+        << "  " << filename << "\n";
+}
+
+// NOTE: is there a better way?
+#ifdef __USE_GNU
+
+std::string check_file(std::string const &filename) {
+    if (filename == "-") { return filename; }
+    struct stat sb;
+    if (stat(filename.c_str(), &sb) != -1) { 
+        if ((sb.st_mode & S_IFMT) == S_IFIFO) { 
+            return filename;
+        }
+        else {
+            std::unique_ptr<char, Free> x(canonicalize_file_name(filename.c_str()));
+            if (x) { return x.get(); }
+        }
+    }
+    return "";
+}
+
+std::pair<std::string, std::string> check_file(std::string const &filename, std::string const &source) {
+    struct stat sb;
+    if (stat(filename.c_str(), &sb) != -1) {
+        if ((sb.st_mode & S_IFMT) == S_IFIFO) {
+            return {filename, filename};
+        }
+        else {
+            std::unique_ptr<char, Free> x(canonicalize_file_name(filename.c_str()));
+            if (x) { return {x.get(), filename}; }
+        }
+    }
+    else if (filename.compare(0, 1, "/", 1) != 0) {
+        std::unique_ptr<char, Free> x(strdup(source.c_str()));
+        std::string path = dirname(x.get());
+        path.push_back('/');
+        path.append(filename);
+        if (stat(path.c_str(), &sb) != -1) { 
+            if ((sb.st_mode & S_IFMT) == S_IFIFO) { 
+                return {path, path};
+            }
+            else {
+                x.reset(canonicalize_file_name(path.c_str()));
+                if (x) { return {x.get(), path}; }
+            }
+        }
+    }
+    return {"", ""};
+}
+
+#else
+
+std::string check_file(std::string const &filename) {
+    if (filename == "-" && std::cin.good()) {
+        return filename; 
+    }
+    if (std::ifstream(filename).good()) {
+        return filename;
+    }
+    return "";
+}
+
+std::pair<std::string, std::string> check_file(std::string const &filename, std::string const &source) {
+    if (std::ifstream(filename).good()) {
+        return {filename, filename};
+    }
+    else {
+#if defined _WIN32 || defined __WIN32__ || defined __EMX__ || defined __DJGPP__
+        const char *SLASH = "\\";
+#else
+        const char *SLASH = "/";
+#endif
+        size_t slash = source.find_last_of(SLASH);
+        if (slash != std::string::npos) {
+            std::string path = source.substr(0, slash + 1);
+            path.append(filename);
+            if (std::ifstream(path).good()) {
+                return {path, path};
+            }
+        }
+    }
+    return {"", ""};
+}
+
+#endif
+
+}
+
 // {{{ defintion of NonGroundParser
 
 NonGroundParser::NonGroundParser(INongroundProgramBuilder &pb)
@@ -63,24 +176,22 @@ void NonGroundParser::pop() { LexerState::pop(); }
 FWString NonGroundParser::filename() const { return LexerState::data().first; }
 
 void NonGroundParser::pushFile(std::string &&file) {
-    auto res = filenames_.insert(std::forward<std::string>(file));
-    if (!res.second) {
-        GRINGO_REPORT(W_FILE_INCLUDED) << "<cmd>: warning: already included file:\n"
-            << "  " << *res.first << "\n";
+    auto checked = check_file(file);
+    if (!checked.empty() && !filenames_.insert(checked).second) {
+        report_included("<cmd>", file);
     }
-    if (!push(*res.first)) {
-        GRINGO_REPORT(ERROR) << "<cmd>: error: '" << *res.first << "' file could not be opened\n";
+    else if (checked.empty() || !push(file)) {
+        report_not_found("<cmd>", file);
     }
 }
 
 void NonGroundParser::pushStream(std::string &&file, std::unique_ptr<std::istream> in) {
     auto res = filenames_.insert(std::move(file));
     if (!res.second) {
-        GRINGO_REPORT(W_FILE_INCLUDED) << "<cmd>: warning: already included file:\n"
-            << "  " << *res.first << "\n";
+        report_included("<cmd>", *res.first);
     }
-    if (!push(*res.first, std::move(in))) {
-        GRINGO_REPORT(ERROR) << "<cmd>: error: '" << *res.first << "' file could not be opened\n";
+    else if (!push(*res.first, std::move(in))) {
+        report_not_found("<cmd>", *res.first);
     }
 }
 
@@ -119,9 +230,12 @@ int NonGroundParser::lex(void *pValue, Location &loc) {
 
 void NonGroundParser::include(unsigned sUid, Location const &loc, bool inbuilt) {
     if (inbuilt) {
-        if (sUid == FWString("iclingo").uid()) {
-            push("<iclingo>", make_unique<std::istringstream>(
-R"(
+        if (sUid == FWString("incmode").uid()) {
+            if (incmodeIncluded_) {
+                report_included(loc, "<incmode>");
+            }
+            else {
+                push("<incmode>", make_unique<std::istringstream>(R"(
 #script (lua) 
 
 function get(val, default)
@@ -133,69 +247,49 @@ function get(val, default)
 end
 
 function main(prg)
-    imin   = get(prg:getConst("imin"), 1)
-    imax   = prg:getConst("imax")
-    istop  = get(prg:getConst("istop"), "SAT")
-    iquery = get(prg:getConst("iquery"), 1)
-    step   = 1
+    local imin   = get(prg:get_const("imin"), 0)
+    local imax   = prg:get_const("imax")
+    local istop  = get(prg:get_const("istop"), "SAT")
 
-    parts = {}
-    parts[#parts+1] = {"base", {}}
-    while true do
-        if imax ~= nil and step > imax then break end
-        parts[#parts+1] = {"cumulative", {step}}
-        ret = gringo.SolveResult.UNKNOWN
-        if step >= iquery then
-            parts[#parts+1] = {"volatile", {step}}
+    local step, ret = 0, gringo.SolveResult.UNKNOWN
+    while (imax == nil or step < imax) and
+          (step == 0   or step < imin or (
+              (istop == "SAT"     and ret ~= gringo.SolveResult.SAT) or
+              (istop == "UNSAT"   and ret ~= gringo.SolveResult.UNSAT) or 
+              (istop == "UNKNOWN" and ret ~= gringo.SolveResult.UNKNOWN))) do
+        local parts = {}
+        table.insert(parts, {"check", {step}})
+        if step > 0 then
+            prg:release_external(gringo.Fun("query", {step-1}))
+            table.insert(parts, {"step", {step}})
+        else
+            table.insert(parts, {"base", {}})
         end
         prg:ground(parts)
-        prg:releaseExternal(gringo.Fun("query", {step-1}))
-        if step >= iquery then
-            prg:assignExternal(gringo.Fun("query", {step}), true)
-            parts = {}
-        end
-        ret = prg:solve()
-        if step >= imin and ((istop == "SAT" and ret == gringo.SolveResult.SAT) or (istop == "UNSAT" and ret == gringo.SolveResult.UNSAT)) then
-            break
-        end
-        step = step+1
+        prg:assign_external(gringo.Fun("query", {step}), true)
+        ret, step = prg:solve(), step+1
     end
 end
+
 #end.
-)"
-            ));
+
+#program check(t).
+#external query(t).
+)"));
+                incmodeIncluded_ = true;
+            }
         }
         else {
-            GRINGO_REPORT(ERROR) << loc << ": error: '" << *FWString(sUid) << "' file could not be opened\n";
+            report_not_found(loc, "<" + *FWString(sUid) + ">");
         }
     }
     else {
-        auto res = filenames_.insert(*FWString(sUid));
-        if (!res.second) {
-            GRINGO_REPORT(W_FILE_INCLUDED) << loc << ": warning: already included file:\n"
-                << "  " << *res.first << "\n";
+        auto paths = check_file(*FWString(sUid), *loc.beginFilename);
+        if (!paths.first.empty() && !filenames_.insert(paths.first).second) {
+            report_included(loc, *FWString(sUid));
         }
-        if (!push(*res.first, true)) {
-#if defined _WIN32 || defined __WIN32__ || defined __EMX__ || defined __DJGPP__
-            const char *SLASH = "\\";
-#else
-            const char *SLASH = "/";
-#endif
-            size_t slash = (*loc.beginFilename).find_last_of(SLASH);
-            if (slash != std::string::npos) {
-                std::string path = (*loc.beginFilename).substr(0, slash + 1);
-                auto res2 = filenames_.insert(path + *res.first);
-                if (!res2.second) {
-                    GRINGO_REPORT(W_FILE_INCLUDED) << loc << ": warning: already included file:\n"
-                        << "  " << *res2.first << "\n";
-                }
-                if (!push(*res2.first, true)) {
-                    GRINGO_REPORT(ERROR) << loc << ": error: '" << *res2.first << "' file could not be opened\n";
-                }
-            }
-            else {
-                GRINGO_REPORT(ERROR) << loc << ": error: '" << *res.first << "' file could not be opened\n";
-            }
+        else if (paths.first.empty() || !push(paths.second, true)) {
+            report_not_found(loc, *FWString(sUid));
         }
     }
 }
