@@ -20,6 +20,8 @@
 
 #include "gringo/output/output.hh"
 #include "gringo/logger.hh"
+#include "gringo/output/aggregates.hh"
+#include <cstring>
 
 namespace Gringo { namespace Output {
 
@@ -169,12 +171,20 @@ struct DefaultLparseTranslator : LparseTranslator {
     using ConstraintVec   = std::vector<LinearConstraint>;
     using DisjointConsVec = std::vector<DisjointConstraint>;
 
-    DefaultLparseTranslator(StmPrinter const &printer)
-        : printer(printer) { }
+    DefaultLparseTranslator(PredDomMap &domains, StmPrinter const &printer)
+        : domains(domains)
+        , printer(printer) { }
     BoundMap::value_type &addBound(Value x) {
         auto it = boundMap.find(x);
         return it != boundMap.end() ? *it : *boundMap.emplace_back(x).first;
     }
+    virtual bool isAtomFromPreviousStep(ULit const &lit) {
+        PredicateDomain::element_type *atom = lit->isAtom();
+        if (!atom) { return false; }
+        if (!atom->first.hasSig()) { return false; }
+        auto it = domains.find(atom->first.sig());
+        return it != domains.end() && atom->second.generation() < it->second.exports.incOffset;
+    }
     static int min() { return std::numeric_limits<int>::min(); }
     static int max() { return std::numeric_limits<int>::max(); }
     virtual void addLowerBound(Value x, int bound) {
@@ -231,7 +241,7 @@ struct DefaultLparseTranslator : LparseTranslator {
             minimizeChanged_ = false;
         }
     }
-    virtual void outputSymbols(LparseOutputter &out, PredDomMap const &domains, OutputPredicates const &outPreds) {
+    virtual void outputSymbols(LparseOutputter &out, OutputPredicates const &outPreds) {
         std::vector<std::tuple<unsigned, Value, int>> symtab;
         auto show = [&](Bound const &bound, AtomState *cond) -> void {
             std::string const *name = nullptr;
@@ -269,14 +279,14 @@ struct DefaultLparseTranslator : LparseTranslator {
         if (it != domains.end()) {
             for (auto jt = it->second.exports.begin() + it->second.exports.showOffset, je = it->second.exports.end(); jt != je; ++jt) {
                 if ((jt->get().first.args().begin()+1)->num() == 1) {
-                    Value val = *(jt->get().first.args().begin());
+                    Value val = jt->get().first.args().front();
                     auto bound = boundMap.find(val);
                     if (bound != boundMap.end()) {
                         if (!showBound(outPreds, *bound)) { show(*bound, &jt->get().second); }
                     }
                     else {
-                        GRINGO_REPORT(W_TERM_UNDEFINED) 
-                            << "warning: trying to show constraint variable that does not occur in program:\n"
+                        GRINGO_REPORT(W_ATOM_UNDEFINED) 
+                            << "info: constraint variable does not occur in program:\n"
                             << "  $" << val << "\n";
                     }
                 }
@@ -288,8 +298,8 @@ struct DefaultLparseTranslator : LparseTranslator {
             if (std::get<1>(x) != Signature("", 0) && std::get<2>(x)) {
                 auto it(seenSigs.find(std::get<1>(x)));
                 if (it == seenSigs.end()) {
-                    GRINGO_REPORT(W_TERM_UNDEFINED) 
-                        << std::get<0>(x) << ": warning: no matching occurrence for signature:\n"
+                    GRINGO_REPORT(W_ATOM_UNDEFINED) 
+                        << std::get<0>(x) << ": info: no constraint variables over signature occur in program:\n"
                         << "  $" << *std::get<1>(x) << "\n";
                     seenSigs.emplace(std::get<1>(x));
                 }
@@ -302,320 +312,8 @@ struct DefaultLparseTranslator : LparseTranslator {
                 << std::get<1>(y)
                 << "="
                 << std::get<2>(y);
-            out.printSymbol(std::get<0>(y), oss.str());
+            out.printSymbol(std::get<0>(y), Value::createId(oss.str()));
         }
-    }
-    void atoms(int atomset, std::function<bool(unsigned)> const &isTrue, ValVec &atoms, PredDomMap const &domains, OutputPredicates const &outPreds) {
-        auto showCsp = [&isTrue, &atoms](Bound const &bound) {
-            assert(!bound.atoms.empty());
-            int prev = bound.atoms.front().first;
-            for (auto it = bound.atoms.begin()+1; it != bound.atoms.end() && !isTrue(it->second->uid); ++it);
-            atoms.emplace_back(Value("$", {bound.var, prev}));
-        };
-        if (atomset & (Model::CSP | Model::SHOWN)) {
-            for (auto &x : boundMap) {
-                if (atomset & Model::CSP || (atomset & Model::SHOWN && showBound(outPreds, x))) { showCsp(x); }
-            }
-        }
-        if (atomset & Model::SHOWN) {
-            auto it(domains.find(Signature("#show", 2)));
-            if (it != domains.end()) {
-                for (auto &y: it->second.domain) {
-                    if (y.first.args().back() == Value(1) && y.second.defined() && y.second.hasUid() && isTrue(y.second.uid())) {
-                        auto bound = boundMap.find(*y.first.args().begin());
-                        if (bound != boundMap.end() && !showBound(outPreds, *bound)) { showCsp(*bound); }
-                    }
-                }
-            }
-        }
-    }
-    void translateMinimize();
-
-    virtual ~DefaultLparseTranslator() { }
-
-    BoundMap           boundMap;
-    ConstraintVec      constraints;
-    DisjointConsVec    disjointCons;
-    MinimizeList       minimize;
-    StmPrinter         printer;
-    unsigned           auxUid = 0;
-    bool               minimizeChanged_ = false;
-    std::set<FWSignature> seenSigs;
-    BoundMap::iterator incBoundOffset;
-};
-
-// }}}
-
-// {{{ definition of DefaultLparseTranslator
-
-void DefaultLparseTranslator::translateMinimize() {
-    // TODO: should detect if there are changes in the minimize constraint
-    // priority -> tuple -> [[lit]]
-    using GroupByPriority = std::map<Value, unique_list<std::pair<FWValVec, std::vector<ULitVec>>, extract_first<FWValVec>>>;
-    GroupByPriority groupBy;
-    for (auto &y : minimize) {
-        auto &ret(groupBy.insert(GroupByPriority::value_type(*(y.first.begin() + 1), GroupByPriority::mapped_type())).first->second);
-        ret.emplace_back(
-            std::piecewise_construct,
-            std::forward_as_tuple(y.first),
-            std::forward_as_tuple()).first->second.emplace_back(get_clone(y.second));
-    }
-    for (auto &y : groupBy) {
-        ULitWeightVec weightLits;
-        for (auto &conds : y.second) {
-            ULitVec condLits;
-            int weight((*conds.first.begin()).num());
-            for (auto &cond : conds.second) {
-                if (cond.size() == 1) {
-                    ULit lit = std::move(cond.front());
-                    condLits.emplace_back(std::move(lit));
-                }
-                else {
-                    // TODO: a function that creates an inverted literal would be nice!!!
-                    SAuxAtom atom(std::make_shared<AuxAtom>(auxAtom()));
-                    printer(LparseRule(atom, std::move(cond)));
-                    condLits.emplace_back(make_unique<AuxLiteral>(atom, false));
-                }
-            }
-            if (condLits.size() == 1) {
-                if (weight < 0) {
-                    SAuxAtom atom(std::make_shared<AuxAtom>(auxAtom()));
-                    printer(LparseRule(atom, nullptr, std::move(condLits.front())));
-                    weightLits.emplace_back(std::move(make_unique<AuxLiteral>(atom, true)), std::abs(weight));
-                }
-                else {
-                    weightLits.emplace_back(std::move(condLits.front()), std::abs(weight));
-                }
-            }
-            else {
-                SAuxAtom atom(std::make_shared<AuxAtom>(auxAtom()));
-                for (auto &cond : condLits) { printer(LparseRule(atom, std::move(cond), nullptr)); }
-                weightLits.emplace_back(make_unique<AuxLiteral>(atom, weight < 0), std::abs(weight));
-            }
-        }
-        printer(LparseMinimize(y.first, std::move(weightLits)));
-    }
-}
-// }}}
-// {{{ definition of Bound
-
-bool Bound::init(DefaultLparseTranslator &x) {
-    if (modified) {
-        modified = false;
-        if (_range.empty()) { x.printer(LparseRule(SAuxAtomVec(), ULitVec(), false)); }
-        else {
-            if (_range.front() == std::numeric_limits<int>::min() || _range.back()+1 == std::numeric_limits<int>::max()) {
-                if      (_range.front()  != std::numeric_limits<int>::min()) { _range.remove(_range.front()+1, std::numeric_limits<int>::max()); }
-                else if (_range.back()+1 != std::numeric_limits<int>::max()) { _range.remove(std::numeric_limits<int>::min(), _range.back()); }
-                else                                                        { _range.clear(), _range.add(0, 1); }
-                GRINGO_REPORT(W_TERM_UNDEFINED)
-                    << "warning: unbounded constraint variable, domain is set to [" << _range.front() << "," << _range.back() << "]:\n"
-                    << "  " << var << "\n";
-                    ;
-            }
-            if (atoms.empty()) {
-                auto assign = [&](SAuxAtom a, SAuxAtom b) {
-                    if (b) {
-                        SAuxAtomVec head;
-                        ULitVec body;
-                        head.emplace_back(b);
-                        if (a) { body.emplace_back(make_unique<AuxLiteral>(a, false)); }
-                        x.printer(LparseRule(std::move(head), std::move(body), true));
-                    }
-                };
-                for (auto y : _range) { 
-                    if (y == _range.front()) { atoms.emplace_back(y, nullptr); } 
-                    else                     { atoms.emplace_back(y, std::make_shared<AuxAtom>(x.auxAtom())); }
-                }
-                for (auto jt = atoms.begin() + 1; jt != atoms.end(); ++jt) { assign(jt->second, (jt-1)->second); }
-                assign(nullptr, atoms.back().second);
-            }
-            else { // incremental update of bounds
-                atom_vec next;
-                int l = _range.front(), r = _range.back();
-                for (auto jt = atoms.begin() + 1; jt != atoms.end(); ++jt) {
-                    int w = (jt - 1)->first;
-                    if (w < l)                           { x.printer(LparseRule(make_unique<AuxLiteral>(jt->second, false), nullptr)); }
-                    else if (w >= r)                     { 
-                        x.printer(LparseRule(make_unique<AuxLiteral>(jt->second, true), nullptr));
-                        if (w == r) { next.emplace_back(*(jt - 1)); }
-                    }
-                    else if (!_range.contains(w, w + 1)) { x.printer(LparseRule(make_unique<AuxLiteral>((jt-1)->second, true), make_unique<AuxLiteral>(jt->second, false))); }
-                    else                                 { next.emplace_back(*(jt - 1)); }
-                }
-                if (atoms.back().first <= r) { next.emplace_back(atoms.back()); }
-                next.front().second = nullptr;
-                atoms = std::move(next);
-            }
-        }
-    }
-    return !_range.empty();
-}
-
-// }}}
-// {{{ definition of LinearConstraint
-
-int depth = 0;
-void LinearConstraint::Generate::generate(StateVec::iterator it, int current, int remainder) {
-    depth++;
-    if (current + remainder > getBound()) {
-        if (it == states.end()) {
-            assert(remainder == 0);
-            aux.emplace_back(std::make_shared<AuxAtom>(trans.auxAtom()));
-            for (auto &x : states) {
-                if (x.atom != x.bound.atoms.end() && x.atom->second) { 
-                    trans.printer(LparseRule(aux.back(), make_unique<AuxLiteral>(x.atom->second, x.coef < 0), nullptr));
-                }
-            }
-        }
-        else {
-            remainder -= it->upper() - it->lower();
-            int total = current - it->lower();
-            for (it->reset(); it->valid(); it->next(total, current)) { 
-                generate(it+1, current, remainder);
-                if (current > getBound()) { break; }
-            }
-        }
-    }
-    depth--;
-}
-
-bool LinearConstraint::Generate::init() {
-    int current   = 0;
-    int remainder = 0;
-    for (auto &y : cons.coefs) {
-        states.emplace_back(*trans.boundMap.find(y.second), y);
-        current   += states.back().lower();
-        remainder += states.back().upper() - states.back().lower();
-    }
-    if (current <= getBound()) {
-        generate(states.begin(), current, remainder); 
-        ULitVec body;
-        for (auto &x : aux) { body.emplace_back(make_unique<AuxLiteral>(x, false)); }
-        trans.printer(LparseRule(cons.atom, std::move(body)));
-    }
-    return current <= cons.bound;
-}
-
-bool LinearConstraint::encode(DefaultLparseTranslator &x) {
-    Generate gen(*this, x);
-    return gen.init();
-}
-
-// }}}
-// {{{ definition of DisjointConstraint
-
-bool DisjointConstraint::encode(DefaultLparseTranslator &x) {
-    std::set<int> values;
-    std::vector<std::map<int, SAuxAtom>> layers;
-    for (auto &elem : cons) {
-        layers.emplace_back();
-        auto &current = layers.back();
-        auto encodeSingle = [&current, &x, &values](ULitVec const &cond, int coef, Value var, int fixed) -> void {
-            auto &bound = *x.boundMap.find(var);
-            auto it = bound.atoms.begin() + 1;
-            for(auto i : bound) {
-                auto &aux = current[i*coef + fixed];
-                if (!aux) { aux = std::make_shared<AuxAtom>(x.auxAtom()); }
-                ULitVec lits = get_clone(cond);
-                if ((it-1)->second)          { lits.emplace_back(make_unique<AuxLiteral>((it-1)->second, true)); }
-                if (it != bound.atoms.end()) { lits.emplace_back(make_unique<AuxLiteral>(it->second, false)); }
-                x.printer(LparseRule(aux, std::move(lits)));
-                values.insert(i*coef + fixed);
-                ++it;
-            }
-        };
-        for (auto &condVal : elem.second) {
-            switch(condVal.value.size()) {
-                case 0: {
-                    auto &aux = current[condVal.fixed];
-                    if (!aux) { aux = std::make_shared<AuxAtom>(x.auxAtom()); }
-                    x.printer(LparseRule(aux, std::move(condVal.lits)));
-                    values.insert(condVal.fixed);
-                    break;
-                }
-                case 1: {
-                    encodeSingle(condVal.lits, condVal.value.front().first, condVal.value.front().second, condVal.fixed);
-                    break;
-                }
-                default: {
-                    // create a bound possibly with holes
-                    auto &b = *x.boundMap.emplace_back(Value("#aux" + std::to_string(x.auxAtom()))).first;
-                    b.clear();
-                    std::set<int> values;
-                    values.emplace(0);
-                    for (auto &mul : condVal.value) {
-                        std::set<int> nextValues;
-                        for (auto i : *x.boundMap.find(mul.second)) {
-                            for (auto j : values) { nextValues.emplace(j + i * mul.first); }
-                        }
-                        values = std::move(nextValues);
-                    }
-                    for (auto i : values) { b.add(i, i+1); }
-                    b.init(x);
-
-                    // create a new variable with the respective bound
-                    // b = value + fixed
-                    // b - value = fixed
-                    // aux :- b - value <=  fixed - 1
-                    // aux :- value - b <= -fixed - 1
-                    // :- aux.
-                    SAuxAtom aux(std::make_shared<AuxAtom>(x.auxAtom()));
-                    condVal.value.emplace_back(-1, b.var);
-                    x.addLinearConstraint(aux, get_clone(condVal.value), -condVal.fixed-1);
-                    for (auto &add : condVal.value) { add.first *= -1; }
-                    x.addLinearConstraint(aux, get_clone(condVal.value), condVal.fixed-1);
-                    x.printer(LparseRule(make_unique<AuxLiteral>(aux, false), nullptr));
-
-                    // then proceed as in case one
-                    encodeSingle(condVal.lits, 1, b.var, 0);
-                    break;
-                }
-            }
-        }
-    }
-    ULitVec checks;
-    for (auto &value : values) {
-        WeightRule::ULitBoundVec card;
-        for (auto &layer : layers) {
-            auto it = layer.find(value);
-            if (it != layer.end()) {
-                card.emplace_back(make_unique<AuxLiteral>(it->second, false), 1);
-            }
-        }
-        SAuxAtom aux = std::make_shared<AuxAtom>(x.auxAtom());
-        checks.emplace_back(make_unique<AuxLiteral>(aux, true));
-        x.printer(WeightRule(aux, 2, std::move(card)));
-    }
-    x.printer(LparseRule(atom, std::move(checks)));
-    return true;
-}
-
-// }}}
-
-// {{{ definition of LparseHandler
-
-struct LparseHandler : StmHandler {
-    LparseHandler(LparseOutputter &out, LparseDebug debug) 
-        : trans([&out, debug](Statement const &x) {
-            if ((unsigned)debug & (unsigned)LparseDebug::LPARSE) { std::cerr << "%%"; x.printPlain(std::cerr); }
-            x.printLparse(out);
-        }) 
-        , out(out)
-        , debug(debug) { }
-    virtual void operator()(Statement &x) { 
-        if ((unsigned)debug & (unsigned)LparseDebug::PLAIN) { std::cerr << "%"; x.printPlain(std::cerr); }
-        x.toLparse(trans);
-    }
-    virtual void incremental() { out.incremental(); }
-    virtual void operator()(PredicateDomain::element_type &head, TruthValue type) {
-        if (!head.second.hasUid()) { head.second.uid(out.newUid()); }
-        out.printExternal(head.second.uid(), type);
-    }
-    virtual void finish(PredDomMap &domains, OutputPredicates &outPreds) {
-        out.disposeMinimize() = trans.minimizeChanged();
-        trans.translate(); 
-        trans.outputSymbols(out, domains, outPreds);
         if (!outPreds.empty()) {
             // show what was requested
             for (auto &x : outPreds) {
@@ -645,19 +343,21 @@ struct LparseHandler : StmHandler {
             }
         }
         // show terms
-        auto it(domains.find(Signature("#show", 2)));
+        it = domains.find(Signature("#show", 2));
         if (it != domains.end()) {
             for (auto jt = it->second.exports.begin() + it->second.exports.showOffset, je = it->second.exports.end(); jt != je; ++jt) {
                 if ((jt->get().first.args().begin()+1)->num() == 0) {
                     if (!jt->get().second.hasUid()) { jt->get().second.uid(out.newUid()); }
-                    out.printSymbol(jt->get().second.uid(), *jt->get().first.args().begin());
+                    out.printSymbol(jt->get().second.uid(), jt->get().first.args().front());
                 }
             }
             it->second.exports.showNext();
         }
-        out.finishSymbols();
+
+        // NOTE: otherwise clasp will complain about redefinition errors...
+        trueLit = nullptr;
     }
-    virtual void atoms(int atomset, std::function<bool(unsigned)> const &isTrue, ValVec &atoms, PredDomMap const &domains, OutputPredicates const &outPreds) {
+    void atoms(int atomset, std::function<bool(unsigned)> const &isTrue, ValVec &atoms, OutputPredicates const &outPreds) {
         if (atomset & (Model::ATOMS | Model::SHOWN)) {
             for (auto &x : domains) {
                 Signature sig = *x.first;
@@ -673,11 +373,316 @@ struct LparseHandler : StmHandler {
             auto it(domains.find(Signature("#show", 2)));
             if (it != domains.end()) {
                 for (auto &y: it->second.domain) {
-                    if (y.first.args().back() == Value(0) && y.second.defined() && y.second.hasUid() && isTrue(y.second.uid())) { atoms.emplace_back(y.first.args().front()); }
+                    if (y.first.args().back() == Value::createNum(0) && y.second.defined() && y.second.hasUid() && isTrue(y.second.uid())) { atoms.emplace_back(y.first.args().front()); }
                 }
             }
         }
-        trans.atoms(atomset, isTrue, atoms, domains, outPreds);
+        auto showCsp = [&isTrue, &atoms](Bound const &bound) {
+            assert(!bound.atoms.empty());
+            int prev = bound.atoms.front().first;
+            for (auto it = bound.atoms.begin()+1; it != bound.atoms.end() && !isTrue(it->second->uid); ++it);
+            atoms.emplace_back(Value::createFun("$", {bound.var, Value::createNum(prev)}));
+        };
+        if (atomset & (Model::CSP | Model::SHOWN)) {
+            for (auto &x : boundMap) {
+                if (atomset & Model::CSP || (atomset & Model::SHOWN && showBound(outPreds, x))) { showCsp(x); }
+            }
+        }
+        if (atomset & Model::SHOWN) {
+            auto it(domains.find(Signature("#show", 2)));
+            if (it != domains.end()) {
+                for (auto &y: it->second.domain) {
+                    if (y.first.args().back() == Value::createNum(1) && y.second.defined() && y.second.hasUid() && isTrue(y.second.uid())) {
+                        auto bound = boundMap.find(y.first.args().front());
+                        if (bound != boundMap.end() && !showBound(outPreds, *bound)) { showCsp(*bound); }
+                    }
+                }
+            }
+        }
+    }
+    void translateMinimize();
+
+    ULit makeAux(NAF naf = NAF::POS) {
+        return make_unique<AuxLiteral>(std::make_shared<AuxAtom>(auxAtom()), naf);
+    }
+
+    ULit getTrueLit() {
+        if (!trueLit) {
+            trueLit = makeAux();
+            LRC().addHead(trueLit).toLparse(*this);
+        }
+        return get_clone(trueLit);
+    }
+
+    virtual ~DefaultLparseTranslator() { }
+
+    PredDomMap        &domains;
+    BoundMap           boundMap;
+    ConstraintVec      constraints;
+    DisjointConsVec    disjointCons;
+    MinimizeList       minimize;
+    StmPrinter         printer;
+    unsigned           auxUid = 0;
+    bool               minimizeChanged_ = false;
+    std::set<FWSignature> seenSigs;
+    BoundMap::iterator incBoundOffset;
+    ULit               trueLit;
+};
+
+// }}}
+
+// {{{ definition of DefaultLparseTranslator
+
+void DefaultLparseTranslator::translateMinimize() {
+    // TODO: should detect if there are changes in the minimize constraint
+    // priority -> tuple -> [[lit]]
+    using GroupByPriority = std::map<Value, unique_list<std::pair<FWValVec, std::vector<ULitVec>>, extract_first<FWValVec>>>;
+    GroupByPriority groupBy;
+    for (auto &y : minimize) {
+        auto &ret(groupBy.insert(GroupByPriority::value_type(*(y.first.begin() + 1), GroupByPriority::mapped_type())).first->second);
+        ret.emplace_back(
+            std::piecewise_construct,
+            std::forward_as_tuple(y.first),
+            std::forward_as_tuple()).first->second.emplace_back(get_clone(y.second));
+    }
+    for (auto &y : groupBy) {
+        ULitWeightVec weightLits;
+        for (auto &conds : y.second) {
+            ULitVec condLits;
+            int weight(conds.first.front().num());
+            for (auto &cond : conds.second) {
+                condLits.emplace_back(getEqualClause(*this, std::move(cond), true, false));
+            }
+            ULit lit = getEqualClause(*this, std::move(condLits), false, false);
+            weightLits.emplace_back(weight < 0 ? lit->negateLit(*this) : std::move(lit), std::abs(weight));
+        }
+        printer(LparseMinimize(y.first, std::move(weightLits)));
+    }
+}
+// }}}
+// {{{ definition of Bound
+
+bool Bound::init(DefaultLparseTranslator &x) {
+    if (modified) {
+        modified = false;
+        if (_range.empty()) { LRC().toLparse(x); }
+        else {
+            if (_range.front() == std::numeric_limits<int>::min() || _range.back()+1 == std::numeric_limits<int>::max()) {
+                if      (_range.front()  != std::numeric_limits<int>::min()) { _range.remove(_range.front()+1, std::numeric_limits<int>::max()); }
+                else if (_range.back()+1 != std::numeric_limits<int>::max()) { _range.remove(std::numeric_limits<int>::min(), _range.back()); }
+                else                                                        { _range.clear(), _range.add(0, 1); }
+                GRINGO_REPORT(W_VARIABLE_UNBOUNDED)
+                    << "warning: unbounded constraint variable:\n"
+                    << "  domain of '" << var << "' is set to [" << _range.front() << "," << _range.back() << "]\n"
+                    ;
+            }
+            if (atoms.empty()) {
+                auto assign = [&](SAuxAtom a, SAuxAtom b) {
+                    if (b) {
+                        LRC rule(true);
+                        if (a) { rule.addBody(make_unique<AuxLiteral>(a, NAF::POS)); }
+                        rule.addHead(make_unique<AuxLiteral>(b, NAF::POS)).toLparse(x);
+                    }
+                };
+                for (auto y : _range) { 
+                    if (y == _range.front()) { atoms.emplace_back(y, nullptr); } 
+                    else                     { atoms.emplace_back(y, std::make_shared<AuxAtom>(x.auxAtom())); }
+                }
+                for (auto jt = atoms.begin() + 1; jt != atoms.end(); ++jt) { assign(jt->second, (jt-1)->second); }
+                assign(nullptr, atoms.back().second);
+            }
+            else { // incremental update of bounds
+                atom_vec next;
+                int l = _range.front(), r = _range.back();
+                for (auto jt = atoms.begin() + 1; jt != atoms.end(); ++jt) {
+                    int w = (jt - 1)->first;
+                    if (w < l)                           { LRC().addBody(make_unique<AuxLiteral>(jt->second, NAF::POS)).toLparse(x); }
+                    else if (w >= r)                     { 
+                        LRC().addBody(make_unique<AuxLiteral>(jt->second, NAF::NOT)).toLparse(x);
+                        if (w == r) { next.emplace_back(*(jt - 1)); }
+                    }
+                    else if (!_range.contains(w, w + 1)) { LRC().addBody(make_unique<AuxLiteral>((jt-1)->second, NAF::NOT)).addBody(make_unique<AuxLiteral>(jt->second, NAF::POS)).toLparse(x); }
+                    else                                 { next.emplace_back(*(jt - 1)); }
+                }
+                if (atoms.back().first <= r) { next.emplace_back(atoms.back()); }
+                next.front().second = nullptr;
+                atoms = std::move(next);
+            }
+        }
+    }
+    return !_range.empty();
+}
+
+// }}}
+// {{{ definition of LinearConstraint
+
+int depth = 0;
+void LinearConstraint::Generate::generate(StateVec::iterator it, int current, int remainder) {
+    depth++;
+    if (current + remainder > getBound()) {
+        if (it == states.end()) {
+            assert(remainder == 0);
+            aux.emplace_back(std::make_shared<AuxAtom>(trans.auxAtom()));
+            for (auto &x : states) {
+                if (x.atom != x.bound.atoms.end() && x.atom->second) { 
+                    LRC()
+                        .addHead(make_unique<AuxLiteral>(aux.back(), NAF::POS))
+                        .addBody(make_unique<AuxLiteral>(x.atom->second, x.coef < 0 ? NAF::NOT : NAF::POS))
+                        .toLparse(trans);
+                }
+            }
+        }
+        else {
+            remainder -= it->upper() - it->lower();
+            int total = current - it->lower();
+            for (it->reset(); it->valid(); it->next(total, current)) { 
+                generate(it+1, current, remainder);
+                if (current > getBound()) { break; }
+            }
+        }
+    }
+    depth--;
+}
+
+bool LinearConstraint::Generate::init() {
+    int current   = 0;
+    int remainder = 0;
+    for (auto &y : cons.coefs) {
+        states.emplace_back(*trans.boundMap.find(y.second), y);
+        current   += states.back().lower();
+        remainder += states.back().upper() - states.back().lower();
+    }
+    if (current <= getBound()) {
+        generate(states.begin(), current, remainder); 
+        ULitVec body;
+        for (auto &x : aux) { body.emplace_back(make_unique<AuxLiteral>(x, NAF::POS)); }
+        LRC().addHead(make_unique<AuxLiteral>(cons.atom, NAF::POS)).addBody(std::move(body)).toLparse(trans);
+    }
+    return current <= cons.bound;
+}
+
+bool LinearConstraint::encode(DefaultLparseTranslator &x) {
+    Generate gen(*this, x);
+    return gen.init();
+}
+
+// }}}
+// {{{ definition of DisjointConstraint
+
+bool DisjointConstraint::encode(DefaultLparseTranslator &x) {
+    std::set<int> values;
+    std::vector<std::map<int, SAuxAtom>> layers;
+    for (auto &elem : cons) {
+        layers.emplace_back();
+        auto &current = layers.back();
+        auto encodeSingle = [&current, &x, &values](ULitVec const &cond, int coef, Value var, int fixed) -> void {
+            auto &bound = *x.boundMap.find(var);
+            auto it = bound.atoms.begin() + 1;
+            for(auto i : bound) {
+                auto &aux = current[i*coef + fixed];
+                if (!aux) { aux = std::make_shared<AuxAtom>(x.auxAtom()); }
+                ULitVec lits = get_clone(cond);
+                if ((it-1)->second)          { lits.emplace_back(make_unique<AuxLiteral>((it-1)->second, NAF::NOT)); }
+                if (it != bound.atoms.end()) { lits.emplace_back(make_unique<AuxLiteral>(it->second, NAF::POS)); }
+                LRC().addHead(make_unique<AuxLiteral>(aux, NAF::POS)).addBody(std::move(lits)).toLparse(x);
+                values.insert(i*coef + fixed);
+                ++it;
+            }
+        };
+        for (auto &condVal : elem.second) {
+            switch(condVal.value.size()) {
+                case 0: {
+                    auto &aux = current[condVal.fixed];
+                    if (!aux) { aux = std::make_shared<AuxAtom>(x.auxAtom()); }
+                    LRC().addHead(make_unique<AuxLiteral>(aux, NAF::POS)).addBody(std::move(condVal.lits)).toLparse(x);
+                    values.insert(condVal.fixed);
+                    break;
+                }
+                case 1: {
+                    encodeSingle(condVal.lits, condVal.value.front().first, condVal.value.front().second, condVal.fixed);
+                    break;
+                }
+                default: {
+                    // create a bound possibly with holes
+                    auto &b = *x.boundMap.emplace_back(Value::createId("#aux" + std::to_string(x.auxAtom()))).first;
+                    b.clear();
+                    std::set<int> values;
+                    values.emplace(0);
+                    for (auto &mul : condVal.value) {
+                        std::set<int> nextValues;
+                        for (auto i : *x.boundMap.find(mul.second)) {
+                            for (auto j : values) { nextValues.emplace(j + i * mul.first); }
+                        }
+                        values = std::move(nextValues);
+                    }
+                    for (auto i : values) { b.add(i, i+1); }
+                    b.init(x);
+
+                    // create a new variable with the respective bound
+                    // b = value + fixed
+                    // b - value = fixed
+                    // aux :- b - value <=  fixed - 1
+                    // aux :- value - b <= -fixed - 1
+                    // :- aux.
+                    SAuxAtom aux(std::make_shared<AuxAtom>(x.auxAtom()));
+                    condVal.value.emplace_back(-1, b.var);
+                    x.addLinearConstraint(aux, get_clone(condVal.value), -condVal.fixed-1);
+                    for (auto &add : condVal.value) { add.first *= -1; }
+                    x.addLinearConstraint(aux, get_clone(condVal.value), condVal.fixed-1);
+                    LRC().addBody(make_unique<AuxLiteral>(aux, NAF::POS)).toLparse(x);
+                    // then proceed as in case one
+                    encodeSingle(condVal.lits, 1, b.var, 0);
+                    break;
+                }
+            }
+        }
+    }
+    ULitVec checks;
+    for (auto &value : values) {
+        WeightRule::ULitBoundVec card;
+        for (auto &layer : layers) {
+            auto it = layer.find(value);
+            if (it != layer.end()) {
+                card.emplace_back(make_unique<AuxLiteral>(it->second, NAF::POS), 1);
+            }
+        }
+        SAuxAtom aux = std::make_shared<AuxAtom>(x.auxAtom());
+        checks.emplace_back(make_unique<AuxLiteral>(aux, NAF::NOT));
+        x.printer(WeightRule(aux, 2, std::move(card)));
+    }
+    LRC().addHead(make_unique<AuxLiteral>(atom, NAF::POS)).addBody(std::move(checks)).toLparse(x);
+    return true;
+}
+
+// }}}
+
+// {{{ definition of LparseHandler
+
+struct LparseHandler : StmHandler {
+    LparseHandler(PredDomMap &domains, LparseOutputter &out, LparseDebug debug) 
+        : trans(domains, [&out, debug](Statement const &x) {
+            if ((unsigned)debug & (unsigned)LparseDebug::LPARSE) { std::cerr << "%%"; x.printPlain(std::cerr); }
+            x.printLparse(out);
+        }) 
+        , out(out)
+        , debug(debug) { }
+    virtual void operator()(Statement &x) { 
+        if ((unsigned)debug & (unsigned)LparseDebug::PLAIN) { std::cerr << "%"; x.printPlain(std::cerr); }
+        x.toLparse(trans);
+    }
+    virtual void incremental() { out.incremental(); }
+    virtual void operator()(PredicateDomain::element_type &head, TruthValue type) {
+        if (!head.second.hasUid()) { head.second.uid(out.newUid()); }
+        out.printExternal(head.second.uid(), type);
+    }
+    virtual void finish(OutputPredicates &outPreds) {
+        out.disposeMinimize() = trans.minimizeChanged();
+        trans.translate(); 
+        trans.outputSymbols(out, outPreds);
+        out.finishSymbols();
+    }
+    virtual void atoms(int atomset, std::function<bool(unsigned)> const &isTrue, ValVec &atoms, OutputPredicates const &outPreds) {
+        trans.atoms(atomset, isTrue, atoms, outPreds);
     }
     virtual ~LparseHandler() { }
 
@@ -690,10 +695,10 @@ struct LparseHandler : StmHandler {
 // {{{ definition of LparsePlainHandler
 
 struct LparsePlainHandler : StmHandler {
-    LparsePlainHandler(std::ostream &out)
-        : trans([&out](Statement const &x)                { x.printPlain(out); })
+    LparsePlainHandler(PredDomMap &domains, std::ostream &out)
+        : trans(domains, [&out](Statement const &x) { x.printPlain(out); })
         , out(out) { }
-    virtual void operator()(Statement &x)                 { x.toLparse(trans); }
+    virtual void operator()(Statement &x) { x.toLparse(trans); }
     virtual void operator()(PredicateDomain::element_type &head, TruthValue type) {
         switch (type) {
             case TruthValue::False: { out << "#external " << head.first << ".\n"; break; }
@@ -702,9 +707,9 @@ struct LparsePlainHandler : StmHandler {
             case TruthValue::Open:  { out << "#external " << head.first << "=open.\n"; break; }
         }
     }
-    virtual void finish(PredDomMap &, OutputPredicates &) { trans.translate(); }
-    virtual void atoms(int, std::function<bool(unsigned)> const &, ValVec &, PredDomMap const &, OutputPredicates const &) { }
-    virtual ~LparsePlainHandler()                         { }
+    virtual void finish(OutputPredicates &) { trans.translate(); }
+    virtual void atoms(int, std::function<bool(unsigned)> const &, ValVec &, OutputPredicates const &) { }
+    virtual ~LparsePlainHandler() { }
 
     DefaultLparseTranslator trans;
     std::ostream           &out;
@@ -725,13 +730,13 @@ struct PlainHandler : StmHandler {
             case TruthValue::Open:  { out << "#external " << head.first << "=open.\n"; break; }
         }
     }
-    virtual void finish(PredDomMap &, OutputPredicates &outPreds) { 
+    virtual void finish(OutputPredicates &outPreds) { 
         for (auto &x : outPreds) { 
             if (std::get<1>(x) != Signature("", 0)) { std::cout << "#show " << (std::get<2>(x) ? "$" : "") << *std::get<1>(x) << ".\n"; }
             else                                    { std::cout << "#show.\n"; }
         }
     }
-    virtual void atoms(int, std::function<bool(unsigned)> const &, ValVec &, PredDomMap const &, OutputPredicates const &) { }
+    virtual void atoms(int, std::function<bool(unsigned)> const &, ValVec &, OutputPredicates const &) { }
     std::ostream &out;
 };
 
@@ -826,10 +831,10 @@ PlainLparseOutputter::~PlainLparseOutputter()                               { }
 // {{{ definition of OutputBase
 
 OutputBase::OutputBase(OutputPredicates &&outPreds, std::ostream &out, bool lparse)
-    : handler(lparse ? UStmHandler(new LparsePlainHandler(out)) : UStmHandler(new PlainHandler(out))) 
+    : handler(lparse ? UStmHandler(new LparsePlainHandler(domains, out)) : UStmHandler(new PlainHandler(out))) 
     , outPreds(std::move(outPreds)) { }
 OutputBase::OutputBase(OutputPredicates &&outPreds, LparseOutputter &out, LparseDebug debug) 
-    : handler(make_unique<LparseHandler>(out, debug))
+    : handler(make_unique<LparseHandler>(domains, out, debug))
     , outPreds(std::move(outPreds)) { }
 void OutputBase::output(Value const &val) {
     auto it(domains.find(val.sig()));
@@ -844,7 +849,11 @@ void OutputBase::output(Value const &val) {
 void OutputBase::incremental() {
     handler->incremental();
 }
-void OutputBase::external(PredicateDomain::element_type &head, TruthValue type) {
+void OutputBase::createExternal(PredicateDomain::element_type &head) {
+    head.second.setExternal(true);
+    (*handler)(head, TruthValue::False);
+}
+void OutputBase::assignExternal(PredicateDomain::element_type &head, TruthValue type) {
     (*handler)(head, type);
 }
 void OutputBase::output(UStm &&x) {
@@ -860,7 +869,20 @@ void OutputBase::flush() {
     stms.clear();
 }
 void OutputBase::finish() { 
-    handler->finish(domains, outPreds);
+    if (!outPreds.empty()) {
+        std::move(outPredsForce.begin(), outPredsForce.end(), std::back_inserter(outPreds));
+        outPredsForce.clear();
+    }
+    handler->finish(outPreds);
+    std::vector <Gringo::FWSignature> rm;
+    for (auto &x : domains) {
+        if (std::strncmp((*(*x.first).name()).c_str(), "#d", 2) == 0) {
+            rm.emplace_back(x.first);
+        }
+    }
+    for (auto const &x : rm) {
+        domains.erase(x);
+    }
 } 
 void OutputBase::checkOutPreds() {
     auto le = [](OutputPredicates::value_type const &x, OutputPredicates::value_type const &y) -> bool { 
@@ -877,7 +899,7 @@ void OutputBase::checkOutPreds() {
             auto it(domains.find(std::get<1>(x)));
             if (it == domains.end()) {
                 GRINGO_REPORT(W_ATOM_UNDEFINED) 
-                    << std::get<0>(x) << ": warning: no matching occurrence for signature:\n"
+                    << std::get<0>(x) << ": info: no atoms over signature occur in program:\n"
                     << "  " << *std::get<1>(x) << "\n";
             }
         }
@@ -885,7 +907,7 @@ void OutputBase::checkOutPreds() {
 }
 ValVec OutputBase::atoms(int atomset, std::function<bool(unsigned)> const &isTrue) const {
     Gringo::ValVec ret;
-    handler->atoms(atomset, isTrue, ret, domains, outPreds);
+    handler->atoms(atomset, isTrue, ret, outPreds);
     return ret;
 }
 Gringo::AtomState const *OutputBase::find(Gringo::Value val) const {
